#include <flashinfer/attention/hopper/prefill_sm90.cuh>
#include "single_prefill_sm90_config.inc"

using namespace flashinfer;

namespace flashinfer {

template cudaError_t SinglePrefillWithKVCacheDispatched
    <{{ head_dim_qk }}, {{ head_dim_vo }}, {{ mask_mode }}, /*USE_SLIDING_WINDOW=*/{{ use_sliding_window }}, {{ variant_name }}, Params>(
    Params& params, cudaStream_t stream);

};
